#ifndef ROBOT_HPP
#define ROBOT_HPP

#include "PCube.h"
#include "Order.hpp"
#include "PlotterObject.hpp"
#include <fstream>
#include <stdio.h>
//|----------------------------------------------------------------------------------------------------------------|
//|                                                     Constans                                                   |
//|----------------------------------------------------------------------------------------------------------------|
#define mm 0.001f
#define cm 0.01f
#define m 1.0f
const float deg=3.141593f/180.0f;
const int distance[4]={1,5,10,25};
const float angle[4]={1.0,5.0,15.0,30.0};
const float max_coord[5]={-1.35f , 1.30f , 0.20f , 90.0f*deg , 0.3f};
const char* filename="log.txt";
//|----------------------------------------------------------------------------------------------------------------|
//|                                            Robot class implementation                                          |
//|----------------------------------------------------------------------------------------------------------------|
class Robot{
	private:
		PlotterStage* stage;
		float coord[5];
		int cubes; 
		bool allowed;
		std::ofstream file;
	private:
		Order* Move(int id,bool direction,int index); 
		Order* goHome(int id);
		Order* goHomeAll();
		Order* Halt(int id);
		Order* HaltAll();
		Order* GetOrder(int id);
	public:
		Robot();
		~Robot();
		float Get(int id); 
		bool IsMoving(int id);
		bool IsMoving();
		Order* Exec(Order* o);
		void Position(int x,int y);
		void PickUp(float x,float y,float d);
		void Transport(PlotterObject obj,int x,int y,int r);
};

//|----------------------------------------------------------------------------------------------------------------|
//|                                              Robot class functions                                             |
//|----------------------------------------------------------------------------------------------------------------|


//---------------------------------------------------Constructor----------------------------------------------------
Robot::Robot(){
	file.open(filename,std::ios_base::app);
	file << (const char*)(System::Runtime::InteropServices::Marshal::StringToHGlobalAnsi(System::DateTime::UtcNow.ToString())).ToPointer()<< " INIT PLOTTERSTAGE" <<std::endl;
	allowed=false;
	try {
		stage = new PlotterStage();
		allowed=true;
	}catch(void* ex){ 
		allowed=false;
		ex;
		throw new int(0);
	}
	cubes=stage->Count();
	for (int i=0;i<cubes;++i){
		coord[i]=stage->getActPos(i);
	}
}

Robot::~Robot(){
	file << (const char*)(System::Runtime::InteropServices::Marshal::StringToHGlobalAnsi(System::DateTime::UtcNow.ToString())).ToPointer()<< " SHUT DOWN PLOTTERSTAGE" <<std::endl;
	if (!allowed) return;
	delete stage;
}

Order* Robot::goHome(int id){
	file << (const char*)(System::Runtime::InteropServices::Marshal::StringToHGlobalAnsi(System::DateTime::UtcNow.ToString())).ToPointer()<< " -PowerCube"<<id<<" is Homed" <<std::endl;
	if (!allowed) return new Order("OK");
	if (id>=cubes || id<0) return new Order("VA");
	stage->goHome(id);
	coord[id]=0;
	return new Order("OK");
}

Order* Robot::goHomeAll(){
	file << (const char*)(System::Runtime::InteropServices::Marshal::StringToHGlobalAnsi(System::DateTime::UtcNow.ToString())).ToPointer()<< " -PlotterStage is Homed" <<std::endl;
	if (!allowed) return new Order("OK");
	for (int id=0;id<cubes;++id) {stage->goHome(id); coord[id]=0;}
	return new Order("OK");
}

Order* Robot::Halt(int id){
	if (!allowed) return new Order("OK");
	if (id>=cubes || id<0) return new Order("VA");
	stage->halt(id);
	coord[id]=stage->getActPos(id);
	stage->reset(id);
	return new Order("OK");
}

Order* Robot::HaltAll(){
	file << (const char*)(System::Runtime::InteropServices::Marshal::StringToHGlobalAnsi(System::DateTime::UtcNow.ToString())).ToPointer()<< " -PlotterStage is Halted"<<std::endl;
	if (!allowed) return new Order("OK");
	stage->haltAll();
	for (int i=0;i<cubes;++i) coord[i]=stage->getActPos(i);
	stage->resetAll();
	return new Order("OK");
}

float Robot::Get(int id){
	if (!allowed) return 0;
	if (id>=cubes && id<0) return -1.0;
	if (id==0 || id==1 || id==2 || id==4) return (float)(int)(10000*stage->getActPos(id)+1)/10;
	if (id==3) return (float)(int)(stage->getActPos(id)/deg);
	return -1.0;
}

Order* Robot::GetOrder(int id){
	if (!allowed) return new Order("A0#000000#");
	if (id>=cubes && id<0) return new Order("VA");
	float tmp=Get(id);
	int i=(int)(abs(tmp*10.0));
	std::stringstream ss;
	ss << i;
	std::string data=ss.str();
	if (data.length()>6) data=data.substr(0,6);
	while (data.length()<6) data="0"+data;
	return new Order("A0#"+data+"#");
}

Order* Robot::Move(int id,bool direction,int index){
	file << (const char*)(System::Runtime::InteropServices::Marshal::StringToHGlobalAnsi(System::DateTime::UtcNow.ToString())).ToPointer()<< " -Move ("<< id <<","<<direction<<","<<index<<");"<< std::endl;
	if (!allowed) return new Order("OK");
	if (index<0 || index>=4) return new Order("VA");
	if (id==1 || id==2){
		if (direction){
			coord[id]-=distance[index]*cm;
			if (coord[id]<0) coord[id]=0;
			stage->movePos(id,coord[id]);
		}
		else{
			coord[id]+=distance[index]*cm;
			if (coord[id]>max_coord[id]) coord[id]=max_coord[id];
			stage->movePos(id,coord[id]);
		}
	}
	if (id==0){
		if (direction){
			coord[id]+=distance[index]*cm;
			if (coord[id] > 0) coord[id]=0;
			stage->movePos(0,coord[id]);
		}
		else{
			coord[id]-=distance[index]*cm;
			if (coord[id]<max_coord[id]) coord[id]=max_coord[id];
			stage->movePos(0,coord[id]);
		}
	}
	if (id==3){
		if (direction){
			coord[id]+=angle[index]*deg;
			if (coord[id]>max_coord[id]) coord[id]=max_coord[id];
			stage->movePos(id,coord[id]);
		}
		else{
			coord[id]-=angle[index]*deg;
			if (coord[id]<-max_coord[id]) coord[id]=-max_coord[id];
			stage->movePos(id,coord[id]);
		}
	}
	if (id==4){
		if (direction){
			coord[id]+=distance[index]*mm;
			if (coord[id]>max_coord[id]) coord[id]=max_coord[id];
			stage->movePos(id,coord[id]);
		}
		else{
			coord[id]-=distance[index]*mm;
			if (coord[id]<0) coord[id]=0;
			stage->movePos(id,coord[id]);
		}
	}
	return new Order("OK");
}

bool Robot::IsMoving(){
	if (!allowed) return false;
	for (int i=0;i<cubes;++i) if (IsMoving(i)) return true;
	return false;
}

bool Robot::IsMoving(int id){
	if (!allowed) return true;
	if (id>=cubes) return false;
	unsigned long cube_state=stage->getState(id);
	cube_state/=2048;
	return (cube_state%2==1);
}

Order* Robot::Exec(Order* o){
	Order* ret=new Order("ER");
	if (o->getOrder()=="W0") ret=Move(0,true,o->getParam()); 
	if (o->getOrder()=="M0") ret=Move(0,false,o->getParam());
	if (o->getOrder()=="W1") ret=Move(1,true,o->getParam()); 
	if (o->getOrder()=="M1") ret=Move(1,false,o->getParam());
	if (o->getOrder()=="W2") ret=Move(2,true,o->getParam());
	if (o->getOrder()=="M2") ret=Move(2,false,o->getParam());
	if (o->getOrder()=="W3") ret=Move(3,true,o->getParam()); 
	if (o->getOrder()=="M3") ret=Move(3,false,o->getParam()); 
	if (o->getOrder()=="M4") ret=Move(4,true,o->getParam()); 
	if (o->getOrder()=="W4") ret=Move(4,false,o->getParam());
	if (o->getOrder()=="H0") ret=goHome(0);
	if (o->getOrder()=="H1") ret=goHome(1);
	if (o->getOrder()=="H2") ret=goHome(2);
	if (o->getOrder()=="H3") ret=goHome(3);
	if (o->getOrder()=="H4") ret=goHome(4);
	if (o->getOrder()=="H9") ret=goHomeAll();
	if (o->getOrder()=="S0") ret=Halt(0);
	if (o->getOrder()=="S1") ret=Halt(1);
	if (o->getOrder()=="S2") ret=Halt(2);
	if (o->getOrder()=="S3") ret=Halt(3);
	if (o->getOrder()=="S4") ret=Halt(4);
	if (o->getOrder()=="S9") ret=HaltAll();
	if (o->getOrder()=="G0") ret=GetOrder(0);
	if (o->getOrder()=="G1") ret=GetOrder(1);
	if (o->getOrder()=="G2") ret=GetOrder(2);
	if (o->getOrder()=="G3") ret=GetOrder(3);
	if (o->getOrder()=="G4") ret=GetOrder(4);
	return ret;
}

void Robot::Position(int x, int y){
	if (!allowed) return;
	stage->movePos(0,-x*mm);
	stage->movePos(1, y*mm);
}

void Robot::PickUp(float x,float y,float d){
	if (!allowed) return;
	stage->movePos(2,0*cm); 
	stage->movePos(0,-x*cm);
	stage->movePos(1,y*cm);
	stage->movePos(3,d*deg);
	while (IsMoving(0) || IsMoving(1) || IsMoving(2) || IsMoving(3));
	stage->movePos(2,20*cm);
}

void Robot::Transport(PlotterObject obj,int x,int y,int r){
	if (!allowed) return;
	//Move to Object
	stage->movePos(2,0*mm);
	stage->movePos(0,-obj.getCoord().getX()*mm);
	stage->movePos(1,obj.getCoord().getX()*mm);
	stage->movePos(3,obj.getRotation()*deg);
	stage->movePos(4,30*mm);	
	while (IsMoving(0) || IsMoving(1) || IsMoving(2) || IsMoving(3) || IsMoving(4));

	//Pick UP
	/**/stage->movePos(2,200*mm);
	while (IsMoving(2));
	/**/stage->movePos(4,29*mm);
	while (IsMoving(4));
	stage->movePos(2,0*mm);
	while (IsMoving(2));

	//Move to Destination
	stage->movePos(0,-x*mm);
	stage->movePos(1,y*mm);
	stage->movePos(3,r*deg);
	while (IsMoving(0) || IsMoving(1) || IsMoving(3));
	//Put Down
	/**/stage->movePos(2,200*mm);
	while (IsMoving(2));
	/**/stage->movePos(4,0*mm);
	while (IsMoving(4));
	stage->movePos(2,0*mm);
	while (IsMoving(2));
}


#endif ROBOT